/*!=============================================================================
  ==============================================================================

  \file    SL_user_openGL.c

  \author  Stefan Schaal
  \date    Nov. 2007

  ==============================================================================
  \remarks
  
  This program contains all graphics user functions for a particular simulation.
  
  ============================================================================*/
  
// SL general includes of system headers
#include "SL_system_headers.h"

// openGL headers
#include "GL/freeglut_std.h"
#include "GL/freeglut_ext.h"
#include "GL/glu.h"
#include <X11/Xlib.h>

// user specific headers
#include "utility.h"
#include "utility_macros.h"

#include "SL.h"
#include "SL_openGL.h"
#include "SL_user.h"
#include "SL_common.h"
#include "SL_unix_common.h"
#include "SL_objects.h"
#include "SL_openGL_servo.h"
#include "SL_collect_data.h"
#include "mdefs.h"
#include "SL_kinematics.h"
#include "SL_shared_memory.h"
#include "SL_userGraphics.h"

// global variables 
int print_Hmat = FALSE;

// local variables
  
// local functions
static int  createWindows(void);
static void myDrawGLElement(int num, double length, int flag);
static void drawCOG(void);
static void drawCenterOfPressure(void);
static void drawCOMs(SL_Jstate *state, SL_endeff *eff, SL_Cstate *basec, SL_quat *baseo);

// global functions
void display(void);
void idle(void);

// external functions


/*!*****************************************************************************
*******************************************************************************
\note  init_user_openGL
\date  July 1998
 
\remarks 
 
initializes everything needed for the graphics for this simulation
 
*******************************************************************************
Function Parameters: [in]=input,[out]=output

\param[in]     argc : number of elements in argv
\param[in]     argv : array of argc character strings
 

******************************************************************************/
int
init_user_openGL(int argc, char** argv)

{
  
  int i,j,n;
  int rc;
  int ans;

  // we need the kinematics initialized
  init_kinematics();

  // read objects
  readObjects(config_files[OBJECTS]);

  // assign contact force mappings
#include "LEKin_contact.h"

  // create simulation windows
  if (!createWindows())
    return FALSE;

  return TRUE;
}

/*!*****************************************************************************
*******************************************************************************
\note  createWindows
\date  July 1998
 
\remarks 
 
initializes graphic windows
 
*******************************************************************************
Function Parameters: [in]=input,[out]=output

none
 

******************************************************************************/
static int
createWindows(void)

{

  int i;
  OpenGLWPtr w;
  int height_main = 400;
  int width_main = 400;
  int hspace = 10;
  int vspace = 60;
  
  Display *disp;
  int  screen_num;
  int  display_width;
  int  display_height;

  char string[100];
  char xstring[100];

  double eye[N_CART+1];
  
  // connect to X server using the DISPLAY environment variable
  if ( (disp=XOpenDisplay(NULL)) == NULL ) {
    printf("Cannot connect to X servo %s\n",XDisplayName(NULL));
    exit(-1);
  }
  
  // get screen size from display structure macro 
  screen_num = DefaultScreen(disp);
  display_width = DisplayWidth(disp, screen_num);
  display_height = DisplayHeight(disp, screen_num);
  
  /* get a window structure, initialized with default values */
  w=getOpenGLWindow();
  if (w==NULL)
    return FALSE;

  w->display = display;
  w->idle    = idle;
  w->width   = width_main;
  w->height  = height_main;

  // check for user parameters
  if (read_parameter_pool_string(config_files[PARAMETERPOOL], 
				 "main_window_geometry", string))
    parseWindowSpecs(string, display_width,display_height,xstring, 
		     &(w->x), 
		     &(w->y), 
		     &(w->width),
		     &(w->height));

  if (read_parameter_pool_double_array(config_files[PARAMETERPOOL], 
				       "main_window_camera_pos", N_CART, eye))
    for (i=1; i<=N_CART; ++i)
      w->eye[i] = eye[i];
  
  // finally create the window
  for (i=1; i<=N_CART; ++i)
    w->eye0[i] = w->eye[i];

  if (!createWindow(w))
    return FALSE;

  return TRUE;

}

/*!*****************************************************************************
*******************************************************************************
\note  display
\date  August 7, 1992
   
\remarks 

this function updates the OpenGL graphics

*******************************************************************************
Function Parameters: [in]=input,[out]=output


******************************************************************************/
void 
display(void)

{
  int i;
  static SL_Jstate  *state = joint_sim_state;
  static SL_endeff  *eff   = endeff;
  static SL_Cstate  *basec = &base_state;
  static SL_quat    *baseo = &base_orient;
  GLfloat  objscolor[4]={(float)0.2,(float)0.2,(float)0.2,(float)1.0};
  static double fscale = 0.1;
  static int    firsttime = TRUE;

  if (firsttime) {
    firsttime = FALSE;
    double w;
    if (read_parameter_pool_double(config_files[PARAMETERPOOL],"contact_force_scale",&w))
      fscale = w;
  }

#include "OpenGL.h"

  // the standard display functions for openGL
#include "SL_user_display_core.h"

  drawCenterOfPressure();
  drawCOG();

  drawCOMs(state, eff, basec, baseo);

}

/*!*****************************************************************************
*******************************************************************************
\note  myDrawGLElement
\date  August 7, 1992
   
\remarks 

draws a GL element of a particular length in z direction

*******************************************************************************
Function Parameters: [in]=input,[out]=output

\param[in]     num   : ID number of element (just for info)
\param[in]     length: length of element
\param[in]     flag  : draw joint element (yes, not)


******************************************************************************/
static void  
myDrawGLElement(int num, double length, int flag)

{
  static int firsttime = TRUE;
  static int chest,neck,head,r_shoulder,l_shoulder,r_arm,l_arm,r_forearm,l_forearm;
  static int hand,phalax,phalax_end,r_hip,l_hip,r_upper_thigh,l_upper_thigh;
  static int r_thigh,l_thigh,r_tibia,l_tibia,l_ankle,r_ankle,r_foot,l_foot;
		
  double width=0.01;

  GLfloat  gray[4]={(float)0.8,(float)0.8,(float)0.8,(float)1.0};
  GLfloat  green[4]={(float)0.1,(float)0.5,(float)0.5,(float)1.0};
  GLfloat  red[4]={(float)1.0,(float)0.25,(float)0.25,(float)1.0};
  GLfloat  blue[4]={(float)0.1,(float)0.1,(float)1.0,(float)1.0};
  GLfloat  black[4]={(float)0.,(float)0.,(float)0.0,(float)1.0};
  GLfloat  white[4]={(float)1.,(float)1.,(float)1.0,(float)1.0};
  GLfloat  yellow[4]={(float)1.0,(float)0.8,(float)0.7,(float)1.0};

  int      isphere = 10;

  if (solid)
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  else
    glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );


  if (firsttime) {
    firsttime = FALSE;

    char *lab_root;
    char  fname[300];

    lab_root = getenv("LAB_ROOT");

    sprintf(fname,"%s/nao/3D/Chest.obj",lab_root);
    chest         = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/Neck.obj",lab_root);
    neck          = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/Head.obj",lab_root);
    head          = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/RShoulderPadMobile.obj",lab_root);
    r_shoulder    = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/LShoulderPadMobile.obj",lab_root);
    l_shoulder    = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/RArm.obj",lab_root);
    r_arm         = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/RForeArm.obj",lab_root);
    r_forearm     = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/LArm.obj",lab_root);
    l_arm         = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/LForeArm.obj",lab_root);
    l_forearm     = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/Hand.obj",lab_root);
    hand          = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/Phalax.obj",lab_root);
    phalax        = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/PhalaxEnd.obj",lab_root);
    phalax_end    = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/RHip.obj",lab_root);
    r_hip         = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/LHip.obj",lab_root);
    l_hip         = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/RUpperThigh.obj",lab_root);
    r_upper_thigh = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/LUpperThigh.obj",lab_root);
    l_upper_thigh = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/RThigh.obj",lab_root);
    r_thigh       = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/LThigh.obj",lab_root);
    l_thigh       = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/RShinebone.obj",lab_root);
    r_tibia       = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/LShinebone.obj",lab_root);
    l_tibia       = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/RAnkle.obj",lab_root);
    r_ankle       = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/LAnkle.obj",lab_root);
    l_ankle       = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/RFoot.obj",lab_root);
    r_foot        = displayListFromObjFile(fname,0.001);

    sprintf(fname,"%s/nao/3D/LFoot.obj",lab_root);
    l_foot        = displayListFromObjFile(fname,0.001);

  }

	
  if (flag==1) {
    glTranslated(0.0,0.0,length); 

    if (num==999) {/* the base coordinate system */
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, blue);
      glColor4fv(blue);
    } else {
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
      glColor4fv(gray);
    } 

    if ((num < 107 ||  num > 130) && num != 6 && num != 12)
      glutSolidSphere(1.2*width,isphere,isphere);
    
    glTranslated(0.0,0.0,-length);
  }

  switch (num) {

  case B_HR: 
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glTranslated(0.00,-0.015,0.029);
    glRotated((GLdouble)-90.,(GLdouble)0.,(GLdouble)0.,(GLdouble)1.);
    glRotated((GLdouble)180.,(GLdouble)1.,(GLdouble)0.,(GLdouble)0.);
    glCallList(chest);
    break;

  case R_SAA:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glCallList(r_shoulder);
    break;

  case R_HR:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)-90.,(GLdouble)0.,(GLdouble)0.,(GLdouble)1.);
    glRotated((GLdouble)9.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glCallList(r_arm);
    break;

  case R_WR:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)-90.,(GLdouble)1.,(GLdouble)0.,(GLdouble)0.);
    glCallList(r_forearm);
    break;

  case L_SAA:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)180.,(GLdouble)1.,(GLdouble)0.,(GLdouble)0.);
    glRotated((GLdouble)180.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glCallList(l_shoulder);
    break;

  case L_HR:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)-90.,(GLdouble)0.,(GLdouble)0.,(GLdouble)1.);
    glRotated((GLdouble)-9.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glCallList(l_arm);
    break;

  case L_WR:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)-90.,(GLdouble)1.,(GLdouble)0.,(GLdouble)0.);
    glCallList(l_forearm);
    break;

  case 117:
  case 131:
    glPopMatrix();
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)90.,(GLdouble)1.,(GLdouble)0.,(GLdouble)0.);
    glRotated((GLdouble)90.,(GLdouble)0.,(GLdouble)0.,(GLdouble)1.);
    glCallList(hand);
    glPushMatrix();
    break;

  case 107:
  case 108:
  case 111:
  case 112:
  case 121:
  case 122:
  case 125:
  case 126:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)-90.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glCallList(phalax);
    break;

  case 115:
  case 129:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)-90.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glRotated((GLdouble)180.,(GLdouble)1.,(GLdouble)0.,(GLdouble)0.);
    glCallList(phalax);
    break;

  case 109:
  case 113:
  case 123:
  case 127:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)-90.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glCallList(phalax_end);
    break;

  case 116:
  case 130:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)-90.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glRotated((GLdouble)180.,(GLdouble)1.,(GLdouble)0.,(GLdouble)0.);
    glCallList(phalax_end);
    break;

  case R_HFE:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)90.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glRotated((GLdouble)90.,(GLdouble)1.,(GLdouble)0.,(GLdouble)0.);
    glCallList(r_hip);
    break;

  case R_HAA:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)90.,(GLdouble)0.,(GLdouble)0.,(GLdouble)1.);
    glCallList(r_upper_thigh);
    break;

  case R_KFE:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glCallList(r_thigh);
    break;

  case R_AFE:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)180.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glCallList(r_tibia);
    break;

  case R_AAA:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)180.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glRotated((GLdouble)90.,(GLdouble)0.,(GLdouble)0.,(GLdouble)1.);
    glCallList(r_ankle);
    break;

  case 132:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glCallList(r_foot);
    break;

  case L_HFE:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)-90.,(GLdouble)0.,(GLdouble)0.,(GLdouble)1.);
    glCallList(l_hip);
    break;

  case L_HAA:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)-90.,(GLdouble)0.,(GLdouble)0.,(GLdouble)1.);
    glCallList(l_upper_thigh);
    break;

  case L_KFE:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glCallList(l_thigh);
    break;

  case L_AFE:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)180.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glCallList(l_tibia);
    break;

  case L_AAA:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)180.,(GLdouble)0.,(GLdouble)1.,(GLdouble)0.);
    glRotated((GLdouble)90.,(GLdouble)0.,(GLdouble)0.,(GLdouble)1.);
    glCallList(l_ankle);
    break;

  case 137:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glCallList(l_foot);
    break;

  case B_HN:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glTranslated(0.0,0.0,length);
    glRotated((GLdouble)-90.,(GLdouble)1.,(GLdouble)0.,(GLdouble)0.);
    glCallList(neck);
    break;

  case 142:
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glColor4fv(gray);
    glRotated((GLdouble)90.,(GLdouble)0.,(GLdouble)0.,(GLdouble)1.);
    glRotated((GLdouble)180.,(GLdouble)1.,(GLdouble)0.,(GLdouble)0.);
    glCallList(head);
    break;

  default:

    if (length == 0 || num == 0)
      break;

    glScaled(width/2.,width/2.,length);
    glColor4fv(gray);
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
    glutSolidCylinder(1.0,1.0,isphere,1);
    
  }
}

/*!*****************************************************************************
*******************************************************************************
\note  idle
\date  June 1999
   
\remarks 

The function called by openGL whenever processing time available.

*******************************************************************************
Function Parameters: [in]=input,[out]=output


******************************************************************************/
void
idle(void)

{
  int i;
  struct timeval t;
  static double last_draw_time = 0.0;
  double current_time;


  //-------------------------------------------------------------------------
  // the main functions are the most handy in an include file

#include "SL_user_idle_core.h"

  // update the eye view positions
  //updateEyes();


  //-------------------------------------------------------------------------
  // refresh display

  gettimeofday(&t,NULL);
  current_time = (double) t.tv_sec + ((double)t.tv_usec)/1.e6;

  if (current_time-last_draw_time >= 1./window_update_rate) {
    glutPostRedisplayAll();
    last_draw_time = current_time; 
  }

}

/*!*****************************************************************************
*******************************************************************************
\note  drawCOG
\date  Feb. 2004
   
\remarks 

draws the position of the center of gravity

*******************************************************************************
Function Parameters: [in]=input,[out]=output

******************************************************************************/
static void
drawCOG(void)

{
  int i,j,n;
  GLfloat   color_point[4]={(float)1.0,(float)1.0,(float)0.5,(float)1.0};
  double    radius = 0.01;
  double    sum;

  // draw a blob at the cog
  glPushMatrix();
  glTranslated((GLdouble)cog.x[_X_],(GLdouble)cog.x[_Y_],(GLdouble)cog.x[_Z_]);
  glColor4fv(color_point);
  if (solid)
    glutSolidSphere(radius,10,10);
  else
    glutWireSphere(radius,10,10);

  /* draw vertical line from COG downwards */
  glDisable(GL_LIGHTING); /*to have constant colors */
  glColor4fv(color_point);
  glBegin(GL_LINES);
  glVertex3d(0.0,0.0,0.0);
  glVertex3d(0.0,0.0,-2.0);
  glEnd();
  glEnable(GL_LIGHTING);
  glPopMatrix();

}

static void 
drawCOMs(SL_Jstate *state, SL_endeff *eff, SL_Cstate *basec,
	 SL_quat *baseo) 

{
  int i, j, n;
  GLfloat color_point[4] = { (float) 0.5, (float) 0.2, (float) 0.2,
			     (float) 1.0 };
  GLfloat color_point2[4] = { (float) 1.0, (float) 0.0, (float) 1.0,
			      (float) 1.0 };
  double radius = 0.005;
  double sum;

  MY_MATRIX(Xmcog,0,N_DOFS,1,3);
  MY_MATRIX(Xaxis,0,N_DOFS,1,3);
  MY_MATRIX(Xorigin,0,N_DOFS,1,3);
  MY_MATRIX(Xlink,0,N_LINKS,1,3);
  MY_MATRIX_ARRAY(Ahmat,1,4,1,4,N_LINKS);
  MY_MATRIX_ARRAY(Ahmatdof,1,4,1,4,N_DOFS);

  // compute the link information for this state
  linkInformation(state, basec, baseo, eff, Xmcog, Xaxis, Xorigin, Xlink,
		  Ahmat, Ahmatdof);

  for (i = 0; i <= N_DOFS; i++) {

    // draw a blob at each center of mass of each link
    glPushMatrix();
    glTranslated((GLdouble) (Xmcog[i][_X_] / links[i].m),
		 (GLdouble) (Xmcog[i][_Y_] / links[i].m),
		 (GLdouble) (Xmcog[i][_Z_] / links[i].m));
    glColor4fv(color_point);

    if (solid) {
      glutSolidSphere(radius, 20, 20);
    } else {
      glutWireSphere(radius, 20, 20);
    }

    glPopMatrix();
  }

}

/*!*****************************************************************************
 *******************************************************************************
\note  drawCenterOfPressure
\date  Feb. 2004
   
\remarks 

      draws the position of the center of pressure

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 ******************************************************************************/
static void
drawCenterOfPressure(void)

{
  int i,j,n;
  GLfloat   color_point[4]={(float)1.0,(float)0.0,(float)0.0,(float)1.0};
  double    radius = 0.02;
  double    xcop[N_CART+1];
  double    x[N_CART+1];
  double    sum;

  /* if there are no objects, exit */
  if (objs==NULL)
    return;

  for (j=1; j<=N_CART; ++j)
    xcop[j]=0;
  sum = 1.e-10;

  for (i=1; i<=n_contacts; ++i) { /* loop over all foot contact points */
    
    // check whether there is an active contact
    if (!contacts[i].active || !contacts[i].status)
      continue;
    
    //printf("contact = %d\n",i);
    
    // where is this contact point
    computeContactPoint(&(contacts[i]),link_pos_sim,Alink_sim,x);
    
    // accumulate the cop position as a weighted sum, where the weight is the
    // z component of the force at each contact point
    for (j=1; j<=N_CART; ++j)
      xcop[j] += x[j] * contacts[i].f[_Z_];
    
    sum += contacts[i].f[_Z_];
    
  }
  

  for (j=1; j<=N_CART; ++j)
    xcop[j] /= sum;

  //printf("z-mass = %f\n",sum);
  
  
  // draw a blob at the cop
  glPushMatrix();
  glTranslated((GLdouble)xcop[_X_],(GLdouble)xcop[_Y_],(GLdouble)xcop[_Z_]);
  glColor4fv(color_point);
  if (solid)
    glutSolidSphere(radius,10,10);
  else
    glutWireSphere(radius,10,10);
  glPopMatrix();

}


