/*!=============================================================================
  ==============================================================================

  \file    SL_user_sensor_proc_runix.c

  \author  Stefan Schaal
  \date    July 2010

  ==============================================================================
  \remarks

  performs reading of sensors, translation to units, and sending out
  motor commands. This verion of the functions connects to the real NAO
  robot via UDP sockets

  ============================================================================*/

// SL general includes of system headers
#include "SL_system_headers.h"

// private includes
#include "utility.h"
#include "utility_macros.h"
#include "SL.h"
#include "SL_common.h"
#include "SL_unix_common.h"
#include "SL_user.h"
#include "SL_sensor_proc.h"
#include "SL_shared_memory.h"
#include "SL_motor_servo.h"
#include "udp_interface.h"
#include "SL_man.h"
#include "SL_collect_data.h"
#include "SL_filters.h"

#define TIME_OUT_NS  1000000000
#define TRANSLATION_FILE "Translation.cf"

typedef struct Translation {
  double slope;
  double offset;
} Translation;


// local variables
static double          raw_positions[N_DOFS+1];
static double          raw_desired_positions[N_DOFS+1];
static double          last_raw_positions[N_DOFS+1];
static double          raw_velocities[N_DOFS+1];
static double          raw_misc_sensors[N_MISC_SENSORS+1];
static double          raw_desired_torques[N_DOFS+1];
static Translation     joint_trans_positions[N_DOFS+1];
static Translation     joint_trans_velocities[N_DOFS+1];
static Translation     joint_trans_torques[N_DOFS+1];
static Translation     joint_trans_desired_torques[N_DOFS+1];
static Translation     misc_trans_sensors[N_MISC_SENSORS+1];
static int             no_motor_commands = FALSE;

// external variables
extern int           motor_servo_errors;

// global functions

// local functions
static int  read_translation(char *fname);
static int  send_sim_state(void);
static int  send_misc_sensors(void);
static int  send_des_command(void);
static void read_trans(void);
static void read_sensor_offs(void);
static void translate_sensor_readings(SL_Jstate *joint_raw_state);
static void translate_misc_sensor_readings(double *misc_raw_sensors);
static void compute_base_state(double *misc_raw);
static double highfilt(double raw, Filter *fptr);


// external functions


/*!*****************************************************************************
 *******************************************************************************
\note  init_user_sensor_processing
\date  Nov. 2007
   
\remarks 

          Initializes all user specific sensory processing

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

     none

 ******************************************************************************/
int
init_user_sensor_processing(void)
{
  int i,j;

  // check for no_motor_command flag
  for (i=1; i<global_argc; ++i) {
    printf("%d %s\n",i,global_argv[i]);
    if (strcmp(global_argv[i],"-nc")==0) {
      no_motor_commands = TRUE;
      printf("INFO: SL will not send motor commands to NAO in -nc mode\n");
    }
  }

  if (!init_udp_interface())
    return FALSE;

  // initalizes translation to and from units
  if (!read_translation(TRANSLATION_FILE))
    return FALSE;

  addToMan("readTranslation","re-reads the translation file",read_trans);
  addToMan("readSensorOffsets","re-reads the sensor-offsets file",read_sensor_offs);

  return TRUE;
}

/*!*****************************************************************************
 *******************************************************************************
\note  read_user_sensors
\date  Dec 1997
   
\remarks 

    gets sensor readings from the robot and converts them into standard
    units

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[out]    raw :     the raw sensor readings
 \param[out]    misc_raw: the raw miscelleneous sensor readings

 ******************************************************************************/
int
read_user_sensors(SL_Jstate *raw,double *misc_raw)
{
  int i,j;
  float fbuf[FBUFLEN_IN];
  int   err;
  static int firsttime = TRUE;


  // get real robot sensory data
  if ((j=read_udp_interface(fbuf, &err)) == CBUFLEN_IN) { // sucess of read

    // save last values
    for (i=1; i<=N_DOFS; ++i)
      last_raw_positions[i] = raw_positions[i];

    // sort into the raw joint angle data
    raw_positions[R_SFE] =  fbuf[R_SHOULDER_PITCH+1];
    raw_positions[R_SAA] =  fbuf[R_SHOULDER_ROLL+1];
    raw_positions[R_HR]  =  fbuf[R_ELBOW_YAW+1];
    raw_positions[R_EB]  =  fbuf[R_ELBOW_ROLL+1];
    raw_positions[R_WR]  =  fbuf[R_WRIST_YAW+1];
    raw_positions[R_FING]=  fbuf[R_NHAND+1];

    raw_positions[L_SFE] =  fbuf[L_SHOULDER_PITCH+1];
    raw_positions[L_SAA] =  fbuf[L_SHOULDER_ROLL+1];
    raw_positions[L_HR]  =  fbuf[L_ELBOW_YAW+1];
    raw_positions[L_EB]  =  fbuf[L_ELBOW_ROLL+1];
    raw_positions[L_WR]  =  fbuf[L_WRIST_YAW+1];    
    raw_positions[L_FING]=  fbuf[L_NHAND+1];

    raw_positions[R_FB]  =  fbuf[L_HIP_YAW_PITCH+1];  // note that R and L are coupled and the same
    raw_positions[R_HFE] =  fbuf[R_HIP_PITCH+1];
    raw_positions[R_HAA] =  fbuf[R_HIP_ROLL+1];
    raw_positions[R_KFE] =  fbuf[R_KNEE_PITCH+1];
    raw_positions[R_AFE] =  fbuf[R_ANKLE_PITCH+1];
    raw_positions[R_AAA] =  fbuf[R_ANKLE_ROLL+1];

    raw_positions[L_FB]  =  fbuf[L_HIP_YAW_PITCH+1];
    raw_positions[L_HFE] =  fbuf[L_HIP_PITCH+1];
    raw_positions[L_HAA] =  fbuf[L_HIP_ROLL+1];
    raw_positions[L_KFE] =  fbuf[L_KNEE_PITCH+1];
    raw_positions[L_AFE] =  fbuf[L_ANKLE_PITCH+1];
    raw_positions[L_AAA] =  fbuf[L_ANKLE_ROLL+1];

    raw_positions[B_HR]  =  fbuf[HEAD_YAW+1];
    raw_positions[B_HN]  =  fbuf[HEAD_PITCH+1];


    // create velocities, taking the delay between package into account
    // not that err is in milliseconds, and indicates how much time difference
    // the sensor data had relative to what was expected on a 100Hz clock
    for (i=1; i<=N_DOFS; ++i)
      raw_velocities[i]  = (raw_positions[i]-last_raw_positions[i])/
	((double)(1+(double)err/10.0)/(double)motor_servo_rate);

    // sorty into the misc sensor data
    raw_misc_sensors[B_XDD_IMU]        = -fbuf[ACC_Y+1];
    raw_misc_sensors[B_YDD_IMU]        = -fbuf[ACC_X+1];
    raw_misc_sensors[B_ZDD_IMU]        = -fbuf[ACC_Z+1];
    raw_misc_sensors[B_ADD_IMU]        = -fbuf[GYR_Y+1];
    raw_misc_sensors[B_BDD_IMU]        = fbuf[GYR_X+1];
    raw_misc_sensors[B_A_ANGLE]        = -fbuf[ANGLE_Y+1];
    raw_misc_sensors[B_B_ANGLE]        = fbuf[ANGLE_X+1];
    raw_misc_sensors[LF_COP_X]         = fbuf[L_COP_X+1];
    raw_misc_sensors[LF_COP_Y]         = fbuf[L_COP_Y+1];
    raw_misc_sensors[LF_TOTAL_WEIGHT]  = fbuf[L_TOTAL_WEIGHT+1];
    raw_misc_sensors[RF_COP_X]         = fbuf[R_COP_X+1];
    raw_misc_sensors[RF_COP_Y]         = fbuf[R_COP_Y+1];
    raw_misc_sensors[RF_TOTAL_WEIGHT]  = fbuf[R_TOTAL_WEIGHT+1];
    raw_misc_sensors[L_SONAR_VAL]      = fbuf[L_SONAR+1];
    raw_misc_sensors[R_SONAR_VAL]      = fbuf[R_SONAR+1];


    // translate the raw values into approriate scale and units
    translate_sensor_readings(raw);
    translate_misc_sensor_readings(misc_raw);

    // track errors
    if (firsttime) {
      firsttime = FALSE;
      scd();
    } else
      motor_servo_errors += round((double)err/10.); // error is in milliseconds


  }

  // compute the base state as much as possible from the misc sensors
  compute_base_state(misc_raw);

  // send the sensor readings (which are processed by now) to
  // shared memory for the graphics visualization
  for (i=1; i<=N_DOFS; ++i)
    joint_sim_state[i] = raw[i];

  for (i=1; i<=N_MISC_SENSORS; ++i)
    misc_sim_sensor[i] = misc_raw[i];

  send_sim_state();
  send_misc_sensors();

  if (semGive(sm_motor_servo_sem) == ERROR)
    exit(-1);

  return TRUE;
}

/*!*****************************************************************************
 *******************************************************************************
\note  read_translation
\date  Feb 2010
   
\remarks 

the translation of raw numbers to and from units

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

   \param[in] fname: the filename to read from

 ******************************************************************************/
static void
read_trans(void)
{
  read_translation(TRANSLATION_FILE);
}

static int
read_translation(char *fname)
{
  
  int       i,j;
  int       rc;
  FILE      *in;
  int       error_flag = FALSE;
  char      string[100];


  /* Get the translation info for each joint: we just parse those
     values from the file config/Translation.cf */

  sprintf(string,"%s%s",CONFIG,fname);
  in = fopen(string,"r");
  if (in == NULL) {
    printf("ERROR: Cannot open file >%s<!\n",string);
    return FALSE;
  }

  /* parse all translation variables according to the joint variables */
  for (i=1; i<= n_dofs; ++i) {
    if (!find_keyword(in, &(joint_names[i][0]))) {
      printf("ERROR: Cannot find offset for >%s<!\n",joint_names[i]);
      fclose(in);
      return FALSE;
    }
    rc = fscanf(in,"%lf %lf %lf %lf %lf %lf %lf %lf",
		&joint_trans_positions[i].slope,
		&joint_trans_positions[i].offset,
		&joint_trans_velocities[i].slope,
		&joint_trans_velocities[i].offset,
		&joint_trans_torques[i].slope,
		&joint_trans_torques[i].offset,
		&joint_trans_desired_torques[i].slope,
		&joint_trans_desired_torques[i].offset);
  }

  for (i=1; i<= n_misc_sensors; ++i) {
    if (!find_keyword(in, &(misc_sensor_names[i][0]))) {
      printf("ERROR: Cannot find offset for >%s<!\n",misc_sensor_names[i]);
      fclose(in);
      return FALSE;
    }
    rc = fscanf(in,"%lf %lf",
		&misc_trans_sensors[i].slope,
		&misc_trans_sensors[i].offset);
  }

  fclose(in);

  return TRUE;

}

/*!*****************************************************************************
 *******************************************************************************
\note  translate_sensor_readings
\date  Dec 1997
   
\remarks 

          translate sensors to unit values

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[out]    raw : the raw state

 ******************************************************************************/
static void
translate_sensor_readings(SL_Jstate *joint_raw_state)

{

  int i,j;
  double temp=0;

  for (i=1; i<=N_DOFS; ++i) {

    // translate the position sensors
    joint_raw_state[i].th = (joint_trans_positions[i].offset + 
			     raw_positions[i]) * joint_trans_positions[i].slope;

    // translate the velocity sensors

    joint_raw_state[i].thd = (joint_trans_velocities[i].offset + 
			      raw_velocities[i]) * joint_trans_velocities[i].slope;

    // translate the force sensors
    joint_raw_state[i].load = 0,0; // we have no info on this

    joint_raw_state[i].th   *= pos_polar[i];
    joint_raw_state[i].thd  *= pos_polar[i];
    joint_raw_state[i].load *= load_polar[i];

    // take out the user offset from the position
    joint_raw_state[i].th -= joint_range[i][THETA_OFFSET];

  }

}

/*!*****************************************************************************
 *******************************************************************************
\note  translate_misc_sensor_readings
\date  Dec 1997
   
\remarks 

          translate miscellaneous sensors to unit values

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[out]    raw : the raw state

 ******************************************************************************/
static void
translate_misc_sensor_readings(double *misc_raw_sensor)

{

  int i,j;
  double temp=0;

  for (i=1; i<=N_MISC_SENSORS; ++i) {

    /* translate the position sensors */

    misc_raw_sensor[i] = (misc_trans_sensors[i].offset + 
			     raw_misc_sensors[i]) * 
      misc_trans_sensors[i].slope;

  }

}


/*!*****************************************************************************
 *******************************************************************************
\note  translate_commands
\date  Dec 1997
   
\remarks 

          translate unit commands into raw values

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[in]     commands : the structure containing the commands

 ******************************************************************************/
static void
translate_commands(SL_Jstate *command)
{

  int i,j;
  double temp=0;
  double raw;

  for (i=1; i<=N_DOFS; ++i) {

    temp = command[i].u * load_polar[i];

    raw = temp/joint_trans_desired_torques[i].slope -
      joint_trans_desired_torques[i].offset; 

    raw_desired_torques[i] = raw;


    temp = joint_des_state[i].th * pos_polar[i];

    raw = temp/joint_trans_positions[i].slope -
      joint_trans_positions[i].offset; 

    raw_desired_positions[i] = raw;

  }

}

/*!*****************************************************************************
 *******************************************************************************
\note  send_user_commands
\date  April 1999
   
\remarks 

    translates the commands into raw and sends them to the robot

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[in]     commands : the structure containing the commands

 ******************************************************************************/
int
send_user_commands(SL_Jstate *command)

{
  int i,j;
  float fbuf[FBUFLEN_OUT];
  float val = 1.0;
  static int transient_counter = 0;
  extern int count_no_broadcast;
  
  // translate into raw values
  translate_commands(command);

  // sort into the raw joint angle data
  fbuf[R_SHOULDER_PITCH+1] =  raw_desired_positions[R_SFE];
  fbuf[R_SHOULDER_ROLL+1]  =  raw_desired_positions[R_SAA];
  fbuf[R_ELBOW_YAW+1]      =  raw_desired_positions[R_HR];
  fbuf[R_ELBOW_ROLL+1]     =  raw_desired_positions[R_EB];
  fbuf[R_WRIST_YAW+1]      =  raw_desired_positions[R_WR];
  fbuf[R_NHAND+1]          =  raw_desired_positions[R_FING];

  fbuf[L_SHOULDER_PITCH+1] =  raw_desired_positions[L_SFE];
  fbuf[L_SHOULDER_ROLL+1]  =  raw_desired_positions[L_SAA];
  fbuf[L_ELBOW_YAW+1]      =  raw_desired_positions[L_HR];
  fbuf[L_ELBOW_ROLL+1]     =  raw_desired_positions[L_EB];
  fbuf[L_WRIST_YAW+1]      =  raw_desired_positions[L_WR];
  fbuf[L_NHAND+1]          =  raw_desired_positions[L_FING];

  fbuf[R_HIP_PITCH+1]      =  raw_desired_positions[R_HFE];
  fbuf[R_HIP_ROLL+1]       =  raw_desired_positions[R_HAA];
  fbuf[R_KNEE_PITCH+1]     =  raw_desired_positions[R_KFE];
  fbuf[R_ANKLE_PITCH+1]    =  raw_desired_positions[R_AFE];
  fbuf[R_ANKLE_ROLL+1]     =  raw_desired_positions[R_AAA];

  fbuf[L_HIP_YAW_PITCH+1]  =  raw_desired_positions[L_FB];
  fbuf[L_HIP_PITCH+1]      =  raw_desired_positions[L_HFE];
  fbuf[L_HIP_ROLL+1]       =  raw_desired_positions[L_HAA];
  fbuf[L_KNEE_PITCH+1]     =  raw_desired_positions[L_KFE];
  fbuf[L_ANKLE_PITCH+1]    =  raw_desired_positions[L_AFE];
  fbuf[L_ANKLE_ROLL+1]     =  raw_desired_positions[L_AAA];

  fbuf[HEAD_YAW+1]         =  raw_desired_positions[B_HR];
  fbuf[HEAD_PITCH+1]       =  raw_desired_positions[B_HN];

  j = R_WRIST_YAW+1;

  fbuf[j+R_SHOULDER_PITCH+1] =  fabs(raw_desired_torques[R_SFE]/u_max[R_SFE]);
  fbuf[j+R_SHOULDER_ROLL+1]  =  fabs(raw_desired_torques[R_SAA]/u_max[R_SAA]);
  fbuf[j+R_ELBOW_YAW+1]      =  fabs(raw_desired_torques[R_HR]/u_max[R_HR]);
  fbuf[j+R_ELBOW_ROLL+1]     =  fabs(raw_desired_torques[R_EB]/u_max[R_EB]);
  fbuf[j+R_WRIST_YAW+1]      =  fabs(raw_desired_torques[R_WR]/u_max[R_WR]);
  fbuf[j+R_NHAND+1]          =  fabs(raw_desired_torques[R_FING]/u_max[R_FING]);

  fbuf[j+L_SHOULDER_PITCH+1] =  fabs(raw_desired_torques[L_SFE]/u_max[L_SFE]);
  fbuf[j+L_SHOULDER_ROLL+1]  =  fabs(raw_desired_torques[L_SAA]/u_max[L_SAA]);
  fbuf[j+L_ELBOW_YAW+1]      =  fabs(raw_desired_torques[L_HR]/u_max[L_HR]);
  fbuf[j+L_ELBOW_ROLL+1]     =  fabs(raw_desired_torques[L_EB]/u_max[L_EB]);
  fbuf[j+L_WRIST_YAW+1]      =  fabs(raw_desired_torques[L_WR]/u_max[L_WR]);
  fbuf[j+L_NHAND+1]          =  fabs(raw_desired_torques[L_FING]/u_max[L_FING]);

  fbuf[j+R_HIP_PITCH+1]      =  fabs(raw_desired_torques[R_HFE]/u_max[R_HFE]);
  fbuf[j+R_HIP_ROLL+1]       =  fabs(raw_desired_torques[R_HAA]/u_max[R_HAA]);
  fbuf[j+R_KNEE_PITCH+1]     =  fabs(raw_desired_torques[R_KFE]/u_max[R_KFE]);
  fbuf[j+R_ANKLE_PITCH+1]    =  fabs(raw_desired_torques[R_AFE]/u_max[R_AFE]);
  fbuf[j+R_ANKLE_ROLL+1]     =  fabs(raw_desired_torques[R_AAA]/u_max[R_AAA]);

  fbuf[j+L_HIP_YAW_PITCH+1]  =  fabs(raw_desired_torques[L_FB]/u_max[L_FB]);
  fbuf[j+L_HIP_PITCH+1]      =  fabs(raw_desired_torques[L_HFE]/u_max[L_HFE]);
  fbuf[j+L_HIP_ROLL+1]       =  fabs(raw_desired_torques[L_HAA]/u_max[L_HAA]);
  fbuf[j+L_KNEE_PITCH+1]     =  fabs(raw_desired_torques[L_KFE]/u_max[L_KFE]);
  fbuf[j+L_ANKLE_PITCH+1]    =  fabs(raw_desired_torques[L_AFE]/u_max[L_AFE]);
  fbuf[j+L_ANKLE_ROLL+1]     =  fabs(raw_desired_torques[L_AAA]/u_max[L_AAA]);

  fbuf[j+HEAD_YAW+1]         =  fabs(raw_desired_torques[B_HR]/u_max[B_HR]);
  fbuf[j+HEAD_PITCH+1]       =  fabs(raw_desired_torques[B_HN]/u_max[B_HN]);

  val *= (double)(!no_motor_commands);
  fbuf[j+R_SHOULDER_PITCH+1] =  val;
  fbuf[j+R_SHOULDER_ROLL+1]  =  val;
  fbuf[j+R_ELBOW_YAW+1]      =  val;
  fbuf[j+R_ELBOW_ROLL+1]     =  val;
  fbuf[j+R_WRIST_YAW+1]      =  val;
  fbuf[j+R_NHAND+1]          =  val;

  fbuf[j+L_SHOULDER_PITCH+1] =  val;
  fbuf[j+L_SHOULDER_ROLL+1]  =  val;
  fbuf[j+L_ELBOW_YAW+1]      =  val;
  fbuf[j+L_ELBOW_ROLL+1]     =  val;
  fbuf[j+L_WRIST_YAW+1]      =  val;
  fbuf[j+L_NHAND+1]          =  val;

  fbuf[j+R_HIP_PITCH+1]      =  val;
  fbuf[j+R_HIP_ROLL+1]       =  val;
  fbuf[j+R_KNEE_PITCH+1]     =  val;
  fbuf[j+R_ANKLE_PITCH+1]    =  val;
  fbuf[j+R_ANKLE_ROLL+1]     =  val;

  fbuf[j+L_HIP_YAW_PITCH+1]  =  val;
  fbuf[j+L_HIP_PITCH+1]      =  val;
  fbuf[j+L_HIP_ROLL+1]       =  val;
  fbuf[j+L_KNEE_PITCH+1]     =  val;
  fbuf[j+L_ANKLE_PITCH+1]    =  val;
  fbuf[j+L_ANKLE_ROLL+1]     =  val;

  fbuf[j+HEAD_YAW+1]         =  val;
  fbuf[j+HEAD_PITCH+1]       =  val;

  // send commands to robot, but wait until some filters settle initially
  if (transient_counter > 10) {
    if (write_udp_interface(fbuf))
      ;
    else {
      ++count_no_broadcast;
    }
  } else {
    ++transient_counter;
  }
  
  return TRUE;

}

/*!*****************************************************************************
 *******************************************************************************
\note  receive_sim_state
\date  Nov. 2007
   
\remarks 

        recieves the entire joint_sim_state from shared memory
	

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

     none

 ******************************************************************************/
static int 
receive_sim_state(void)
{
  
  int i;

  if (semTake(sm_joint_sim_state_sem,ns2ticks(TIME_OUT_NS)) == ERROR) {
    
    ++motor_servo_errors;
    return FALSE;

  } 

  for (i=1; i<=n_dofs; ++i)
    sm_joint_sim_state_data[i] = sm_joint_sim_state->joint_sim_state[i];
  
  cSL_Jstate(joint_sim_state,sm_joint_sim_state_data,n_dofs,FLOAT2DOUBLE);

  // get time stamp and check for synchronization errors
  motor_servo_time = servo_time = sm_joint_sim_state->ts;

  semGive(sm_joint_sim_state_sem);

  return TRUE;
}

/*!*****************************************************************************
 *******************************************************************************
\note  receive_misc_sensors
\date  Nov. 2007   
\remarks 

        receives the entire misc_sim_sensors from shared memory
	

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

     none

 ******************************************************************************/
static int 
receive_misc_sensors(void)
{
  
  int i;

  if (n_misc_sensors <= 0)
    return TRUE;

  if (semTake(sm_misc_sim_sensor_sem,ns2ticks(TIME_OUT_NS)) == ERROR) {
    
    ++motor_servo_errors;
    return FALSE;

  } 

  for (i=1; i<=n_misc_sensors; ++i)
    misc_sim_sensor[i] = sm_misc_sim_sensor->value[i];
  
  semGive(sm_misc_sim_sensor_sem);

  return TRUE;
}

/*!*****************************************************************************
 *******************************************************************************
\note  send_des_command
\date  Nov. 2007
   
\remarks 

        send the commands from the joint_sim_state shared memory
        structure
	

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

     none

 ******************************************************************************/
static int 
send_des_command(void)
{
  
  int i;
  extern double *upd;

  if (semTake(sm_des_commands_sem,ns2ticks(TIME_OUT_NS)) == ERROR) {
    
    ++motor_servo_errors;
    return FALSE;

  } 

  for (i=1; i<=n_dofs; ++i) {
    sm_des_commands->des_command[i].th  = (float) joint_des_state[i].th;
    sm_des_commands->des_command[i].thd = (float) joint_des_state[i].thd;
    sm_des_commands->des_command[i].uff = (float) joint_des_state[i].uff;
    sm_des_commands->des_command[i].u   = (float) joint_sim_state[i].u;
    sm_des_commands->des_command[i].upd = (float) upd[i];
  }

  sm_des_commands->ts = motor_servo_time;
  
  semGive(sm_des_commands_sem);

  return TRUE;
}

/*!*****************************************************************************
 *******************************************************************************
\note  user_controller
\date  Feb. 2009
   
\remarks 

          allows the user to modify the commands currently computed by 
          the active controller. Note that only u is actually used in
          the final send out of the commands to the robot, and ufb is only
          to kept as feedback to the user for data collection.

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[in,out] u     : the current total command
 \param[in,out] ufb   : the current feedback command

 ******************************************************************************/
void
user_controller(double *u, double *uf)
{
  int i,j;
}

/*!*****************************************************************************
 *******************************************************************************
\note  userCheckForMessage
\date  Feb. 2009
   
\remarks 

          this function allows the user to intercept a message send to the
          motor servo and use this information in sensor_user_proc

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[in]     name : message identifying name
 \param[in]     k    : index of message in shared memory

 ******************************************************************************/
void
userCheckForMessage(char *name, int k)
{
  int i,j;

}

/*!*****************************************************************************
*******************************************************************************
\note  send_sim_state
\date  Nov. 2007
   
\remarks 

sends the entire joint_sim_state to shared memory


*******************************************************************************
Function Parameters: [in]=input,[out]=output

none

******************************************************************************/
static int 
send_sim_state(void)
{
  
  int i;

  // joint state
  if (semTake(sm_joint_sim_state_sem,ns2ticks(TIME_OUT_NS)) == ERROR) {
    
    ++motor_servo_errors;
    return FALSE;

  } 

  cSL_Jstate(joint_sim_state,sm_joint_sim_state_data,n_dofs,DOUBLE2FLOAT);
    
  for (i=1; i<=n_dofs; ++i)
      sm_joint_sim_state->joint_sim_state[i] = sm_joint_sim_state_data[i];
  
  semGive(sm_joint_sim_state_sem);


  // base state
  if (semTake(sm_base_state_sem,ns2ticks(TIME_OUT_NS)) == ERROR) {
    
    ++motor_servo_errors;
    return FALSE;

  } 

  cSL_Cstate((&base_state)-1, sm_base_state_data, 1, DOUBLE2FLOAT);

  sm_base_state->state[1] = sm_base_state_data[1];

  sm_base_state->ts = motor_servo_time;
  
  semGive(sm_base_state_sem);


  // base orient
  if (semTake(sm_base_orient_sem,ns2ticks(TIME_OUT_NS)) == ERROR) {
    
    ++motor_servo_errors;
    return FALSE;

  } 

  cSL_quat(&base_orient-1, sm_base_orient_data, 1, DOUBLE2FLOAT);

  sm_base_orient->orient[1] = sm_base_orient_data[1];

  sm_base_orient->ts = motor_servo_time;
  
  semGive(sm_base_orient_sem);

  return TRUE;
}

/*!*****************************************************************************
 *******************************************************************************
\note  send_misc_sensors
\date  Feb. 2010
   
\remarks 

sends the entire misc_sim_sensor to shared memory
	

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 none
 
******************************************************************************/
static int 
send_misc_sensors(void)
{
  
  int i;
  
  if (n_misc_sensors <= 0)
    return TRUE;
  
  if (semTake(sm_misc_sim_sensor_sem,ns2ticks(TIME_OUT_NS)) == ERROR) {
    
    ++motor_servo_errors;
    return FALSE;
    
  } 
  
  for (i=1; i<=n_misc_sensors; ++i)
    sm_misc_sim_sensor->value[i] = misc_sim_sensor[i];
  
  semGive(sm_misc_sim_sensor_sem);
  
  return TRUE;
}

/*!*****************************************************************************
 *******************************************************************************
\note  read_sensor_offs
\date  Oct 2011
   
\remarks 

interface to read sensor offsets configuration file

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 none

 ******************************************************************************/
static void
read_sensor_offs(void)
{
  read_sensor_offsets(config_files[SENSOROFFSETS]);
}

/*!*****************************************************************************
 *******************************************************************************
\note  compute_base_state
\date  April 2016
   
\remarks 

 computes all base state variables as much as possible

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 none

 ******************************************************************************/
static void
compute_base_state(double *misc_raw)
{

  int i;
  static double acc_offset[N_CART+1];
  static Filter fptr_acc[N_CART+1];
  static Filter fptr_vel[N_CART+1];
  static int    count=0;
  MY_VECTOR(a,1,N_CART);
  MY_MATRIX(R,1,N_CART,1,N_CART);

  // quaternion of base
  
  // note: the eulerToRotMat function also works for fixed angles by
  // assigning yaw,pitch,roll to elements 1,2,3

  a[_A_] = -misc_raw[B_A_ANGLE];
  a[_B_] = -misc_raw[B_B_ANGLE];
  a[_G_] = 0.0; // yaw not known
  
  eulerToRotMat(a,R);
  linkQuat(R, &base_orient);

  // linear acceleration of body needs to be converted world coordinates
  a[_X_] = misc_raw[B_XDD_IMU];
  a[_Y_] = misc_raw[B_YDD_IMU];
  a[_Z_] = misc_raw[B_ZDD_IMU];

  mat_vec_mult_size(R,N_CART,N_CART,a,N_CART,base_state.xdd);

  if (count <= 100) {
    ++count;
    if (count == 1) {
      for (i=1; i<=N_CART; ++i)
	acc_offset[i] = base_state.xdd[i];
    } else {
      for (i=1; i<=N_CART; ++i)
	acc_offset[i] = (acc_offset[i]*(count-1)+base_state.xdd[i])/(double)(count);
    }
  }

  for (i=1; i<=N_CART; ++i)
    base_state.xdd[i] -= acc_offset[i];
  
  // integrate linear acceleration after high-pass filtering
  for (i=1; i<=N_CART; ++i) {
    //base_state.xd[i] += highfilt(base_state.xdd[i],&(fptr_acc[i]))/(double)(motor_servo_rate);
    base_state.xd[i] = base_state.xdd[i]/(double)(motor_servo_rate) + base_state.xd[i]*1.0;
  }
  for (i=1; i<=N_CART; ++i) {
    //base_state.x[i] += highfilt(base_state.xd[i],&(fptr_vel[i]))/(double)(motor_servo_rate);
    base_state.x[i] = base_state.xd[i]/(double)(motor_servo_rate) + 0.95*base_state.x[i];
    base_state.x[i] = 0;
  }
  
}

/*!*****************************************************************************
 *******************************************************************************
\note  highfilt
\date  April 2016
   
\remarks 

    applies a 2nd order butteworth highpass filter

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[in]     raw : the raw data
 \param[in]     fptr: pointer to the filter data structure 

 returns the filtered value

 ******************************************************************************/
double
highfilt(double raw, Filter *fptr)

{

  int i,j;
  static double fb[] = {0.995566972017647,  -1.991133944035294,   0.995566972017647};
  static double fa[] = {1.000000000000000,  -1.991114292201654,   0.991153595868935};

  fptr->raw[0]  = raw;
  
  fptr->filt[0] = 
    fb[0] * fptr->raw[0] +
    fb[1] * fptr->raw[1] + 
    fb[2] * fptr->raw[2] -
    fa[1] * fptr->filt[1] -
    fa[2] * fptr->filt[2];
  
  fptr->raw[2]  = fptr->raw[1];
  fptr->raw[1]  = fptr->raw[0];
  fptr->filt[2] = fptr->filt[1];
  fptr->filt[1] = fptr->filt[0];

  return fptr->filt[0];

}
