/*!=============================================================================
  ==============================================================================

  \file    udp_interface.cpp

  \author  Stefan Schaal
  \date    Feb. 2016

  ==============================================================================
  \remarks

  Simple interface program to the C++ UDP library for the NAO. The strategy is
  to using the blocking socket for timing for the motor_servo, and to give the
  semaphore for the motor servo out of this program and SL_user_motor.c (only
  during initialization, as the shared memory is not initizlied when 
  init_udp_interface is called.

  Thus, given that the NAO times the socket packages at 100Hz, we realize a
  100 Hz communication with minimal packet loss.

  ============================================================================*/

#include <unistd.h>

// private includes
#include "utility.h"
#include "UDP_communication.h"
#include "udp_interface.h"


using namespace udp_communication;

// the ports need to switched relative to NAO communication, i.e., client
// becomes server, and server becomes client
#define UDPSERVERPORT     55003
#define UDPCLIENTPORT     55002

//union structure for UDP communication for sending all sensor values
static union {
  char   cbuf[CBUFLEN_IN];
  float  fbuf[FBUFLEN_IN];
} in_buf;

static union {
  char   cbuf[CBUFLEN_OUT];
  float  fbuf[FBUFLEN_OUT];
} out_buf;


// local variables
UDP_communication udpServer;
UDP_communication udpClient;
static int last_frame = 0;

// external variables
extern int           motor_servo_errors;

// global functions


// local functions

/*!*****************************************************************************
 *******************************************************************************
\note  init_udp_interface
\date  Feb. 2016
   
\remarks 

          Initializes all user specific sensory processing

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

     none

 ******************************************************************************/
int
init_udp_interface(void)
{
  int i,j;
  char name[100] = "10.0.0.?";
  int  n_bytes_ready = 0;
  int  n_bytes=0;
  int  counter=0;
  char empty[100]="";
  FILE *fp;


  // remember the last used IP address
  fp = fopen(".nao_ip","r");
  if (fp!=NULL) {
    fscanf(fp,"%s",name);
    fclose(fp);
  }

  // need IP address of NAO robot -- just push NAO's chest button
  if (!get_string("\nIP Address of NAO Robot?", name, name)) {
    return FALSE;
  }

  // remember the last used IP address
  fp = fopen(".nao_ip","w");
  if (fp!=NULL) {
    fprintf(fp,"%s",name);
    fclose(fp);
  }

  // initalize the server socket
  if (!udpServer.makeUDPServer(UDPSERVERPORT,empty)) {
    return FALSE;
  }

  // create the client socket to communicate with this NAO robot
  if (!udpClient.makeUDPClient(UDPCLIENTPORT,name)) {
    return FALSE;
  }

  // the first package sent to the NAO will trigger the communication
  // protocol on the robot. The content of this package will be ignored
  for (i=0; i<FBUFLEN_OUT; ++i)
      out_buf.fbuf[i] = 0.0;

  printf("Starting UDP communication to %s.%d ...\n",name,UDPCLIENTPORT);
  printf("Waiting for UDP return package at %s.%d: .",empty,UDPSERVERPORT);
  fflush(stdout);

  while (n_bytes == 0) { // in case we lose a UPD package, run until success

    if (++counter > 25)
      return FALSE;
    
    // send the package
    udpClient.writeUDPSocket(out_buf.cbuf,CBUFLEN_OUT);

    sleep(1); // just a small time out to allow the communication to start

    // see whether we get a package returned and empty the socket buffer
    while ((n_bytes_ready=udpServer.checkUDPSocket())) {
      n_bytes = udpServer.readUDPSocket(in_buf.cbuf,CBUFLEN_IN,NULL);
    }

    if (n_bytes == 0) {
      printf("%d.",counter);
      fflush(stdout);
    }

  }

  // now we should be communicating fine
  printf("\nCommunication to NAO at %s established\n",name);

  // NOTE: at this point we have most likely sent several UDP packages to
  // the client. The client needs to make sure to catch up with too many
  // packages in its socket buffer

  return TRUE;
}


/*!*****************************************************************************
 *******************************************************************************
\note  read_udp_interface
\date  Feb 2016
   
\remarks 

          reads data from the socket

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

     \param [out]     fbuf: float array with FBUFLEN elements, that will
                            contain all sensors starting with index 1, and 
                            the frame counter in the zero-th element

     \param [out]      err: count of missed frames relative to last read

     returns the number of bytes read

 ******************************************************************************/
int
read_udp_interface(float *fbuf, int *err)
{
  int i,j;
  int  n_bytes_ready = 0;
  int  new_frame;
  int  n_bytes=0;
  extern double count_no_receive_total;

  *err = 0;
  new_frame = last_frame;

  // only return if the next frame counter is higher than the current one,
  // otherwise we hang in the blocking socket
  while ( new_frame <= last_frame ) {

    // this call blocks and releases with a new package
    n_bytes = udpServer.readUDPSocket(in_buf.cbuf,CBUFLEN_IN,NULL);
  
    // check the frame counter: we only accept the package if the frame
    // counter increases, thus potentially skipping a badly order UDP package
    // err will reflect such errors, and will be used for fixing numerical
    // differentiation for velocities in SL_user_sensor_runix.c
  
    if (in_buf.fbuf[FBUFLEN_IN-1] > new_frame) {
      new_frame = in_buf.fbuf[FBUFLEN_IN-1];
      fbuf[0] = new_frame;
      for (i=0; i<FBUFLEN_IN-1; ++i)
	fbuf[i+1] = in_buf.fbuf[i];
    } else {
      count_no_receive_total += 1.0;
    }

  }
  
  if (new_frame-last_frame != 10) { // frames should be on 10ms scale
    *err = new_frame-last_frame-10; // will be an average drift
  }

  last_frame = new_frame;

  return n_bytes;
}

/*!*****************************************************************************
 *******************************************************************************
\note  write_udp_interface
\date  Feb 2016
   
\remarks 

          reads data from the socket

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

     \param [in]      fbuf: float array with FBUFLEN elements, that will
                            contain all sensors starting with index 1, and 
                            the frame counter in the zero-th element

     returns TRUE or FALSE

 ******************************************************************************/
int
write_udp_interface(float *fbuf)
{
  int  i,j;
  int  n_bytes=0;

  for (i=0; i<FBUFLEN_OUT-1; ++i)
    out_buf.fbuf[i] = fbuf[i+1];

  // add a frame counter
  out_buf.fbuf[FBUFLEN_OUT-1] = last_frame; // we send the package with the same frame
                                            // counter as the receive package, as write
                                            // comes after read in SL. This allows the
                                            // the client to check for a matching UDP
                                            // package relative to its last sent out
                                            // sensor values
  
  if ((n_bytes=udpClient.writeUDPSocket(out_buf.cbuf,CBUFLEN_OUT)) != CBUFLEN_OUT) {
    printf("Couldn't write all bytes %d != %d\n",n_bytes,(int)CBUFLEN_OUT);
    return FALSE;
  }

  return TRUE;
}

