/*!=============================================================================
  ==============================================================================

  \file    user_common.c

  \author 
  \date   

  ==============================================================================
  \remarks

      useful functions shared by various programs

  ============================================================================*/

#include "SL.h"
#include "SL_user.h"
#include "SL_common.h"
#include "SL_dynamics.h"
#include "math.h"
#include "utility.h"
#include "utility_macros.h"

// local functions
static void
reorthonormalizeMatrix( Matrix A, Matrix B, int nr, int nc, Matrix C);

static void 
computeFullFloatingBaseState(int nu, Matrix Vu, Matrix dVudt, Matrix SVu, double *qdd_ref,
			     double *qd_ref, double *xdd_ref, double *add_ref);



/*!*****************************************************************************
 *******************************************************************************
\note  baseToWorld
\date  April 2006
   
\remarks 

        converts a 3D position in base coordinates to world coordinates
        position

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output


 \param[in]     xl     : the vector in local coordinates
 \param[in]     cbase  : the base position
 \param[in]     obase  : the base orientation
 \param[out]    xw     : the vector in world coordinates


 ******************************************************************************/
void
baseToWorld(double *xl, SL_Cstate *cbase, SL_quat *obase, double *xw)
{

  int           i,j;
  static int    firsttime=TRUE;
  double        tmp[N_CART+1];

  MY_MATRIX(R,1,N_CART,1,N_CART);
  
  quatToRotMatInv(obase,R);
  mat_vec_mult_size(R,N_CART,N_CART,xl,N_CART,tmp);

  for (i=1; i<=N_CART; ++i)
    xw[i] = tmp[i] + cbase->x[i];

}

/*!*****************************************************************************
 *******************************************************************************
\note  worldToBase
\date  April 2006
   
\remarks 

        converts a 3D position in world coordinates to base coordinates

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output


 \param[in]     xw     : the vector in world coordinates
 \param[in]     cbase  : the base position
 \param[in]     obase  : the base orientation
 \param[out]    xl     : the vector in local coordinates


 ******************************************************************************/
void
worldToBase(double *xw, SL_Cstate *cbase, SL_quat *obase, double *xl)
{

  int           i,j;
  static int    firsttime=TRUE;
  double        temp[N_CART+1];

  MY_MATRIX(R,1,N_CART,1,N_CART);

  for (i=1; i<=N_CART; ++i)
    temp[i] = xw[i] - cbase->x[i];

  quatToRotMat(obase,R);
  mat_vec_mult_size(R,N_CART,N_CART,temp,N_CART,xl);

}

/*!*****************************************************************************
 *******************************************************************************
\note  worldToBaseVelocity
\date  April 2006
   
\remarks 

        converts a 3D velocity in world coordinates to base coordinates

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output


 \param[in]     xw     : the position vector in world coordinates
 \param[in]     vw     : the velocity vector in world coordinates
 \param[in]     cbase  : the base position and velocity
 \param[in]     obase  : the base orientation and angular velocity
 \param[out]    vl     : the velocity vector in local coordinates


 ******************************************************************************/
void
worldToBaseVelocity(double *xw, double *vw, 
		    SL_Cstate *cbase, SL_quat *obase, double *vl)
{

  int           i,j;
  static int    firsttime=TRUE;
  double temp1[N_CART+1];
  double temp2[N_CART+1];
  
  MY_MATRIX(R,1,N_CART,1,N_CART);

  // the difference between point of interest and base
  for (i=1; i<=N_CART; ++i)
    temp1[i] = xw[i]-cbase->x[i];

  // velocity term due to angular velocity
  temp2[_X_] = obase->ad[_Y_]*temp1[_Z_] - obase->ad[_Z_]*temp1[_Y_];
  temp2[_Y_] = obase->ad[_Z_]*temp1[_X_] - obase->ad[_X_]*temp1[_Z_];
  temp2[_Z_] = obase->ad[_X_]*temp1[_Y_] - obase->ad[_Y_]*temp1[_X_];

  for (i=1; i<=N_CART; ++i)
    temp1[i] = vw[i] - cbase->xd[i] - temp2[i];

  quatToRotMat(obase,R);
  mat_vec_mult_size(R,N_CART,N_CART,temp1,N_CART,vl);

}


/*!*****************************************************************************
 *******************************************************************************
\note  compute_cog_jacobian
\date  March 2006
   
\remarks 

      computes the center of gravity jacobian

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

     All results can be computed based on global variables. Results are
     assigned to the Jcog matrix


 ******************************************************************************/
void
compute_cog_jacobian(void)
{
  int    i,j,r;
  double mxcog[N_CART+1];
  double mxcog_des[N_CART+1];
  double mz[N_CART+1];
  double mz_des[N_CART+1];
  double morigin_pos[N_CART+1];
  double c[2*N_CART+1];
  double mass_total;
  double mass_contrib;


#include "nao_COG_GJac_declare.h"

#include "nao_COG_GJac_math.h"

  /* the Jlist variable generated by the math files contains the 
     the indicators which joints contribute to the COG jacobian of
     each link. Now, this information is used to compute the
     geometric jacobian of the entire COG */

  // compute the total mass
  mass_total = links[BASE].m;
  for (i=1; i<=N_DOFS; ++i) {
    mass_total += links[i].m;
  }

  mat_zero(Jcog);
  mat_zero(Jcogdes);

  for (i=1; i<=N_DOFS; ++i) {    // accumulate the Jacobians of all links

    for (j=1; j<=N_DOFS; ++j) {  // go over all columns of the Jacobian
    
      if ( Jlist[i][j] == 0 )
	continue;

      for (r=1; r<=N_CART; ++r)
	morigin_pos[r] = links[i].m*joint_origin_pos[j][r];

      revoluteGJacColumn(joint_cog_mpos[i],morigin_pos,joint_axis_pos[j],c );

      // add to Jcog
      for (r=1; r<=2*N_CART; ++r)
	if (r>N_CART) // need to correct for missing mass multiplier
	  Jcog[r][j] += c[r]*links[i].m;    
	else
	  Jcog[r][j] += c[r];    


      for (r=1; r<=N_CART; ++r)
	morigin_pos[r] = links[i].m*joint_origin_pos_des[j][r];

      revoluteGJacColumn(joint_cog_mpos_des[i],morigin_pos,joint_axis_pos_des[j],c );

      // add to Jcogdes
      for (r=1; r<=2*N_CART; ++r)
	if (r>N_CART)  // need to correct for missing mass multiplier
	  Jcogdes[r][j] += c[r]*links[i].m;    
	else
	  Jcogdes[r][j] += c[r];    

    }

  }

  mat_mult_scalar(Jcog,1./mass_total,Jcog);
  mat_mult_scalar(Jcogdes,1./mass_total,Jcogdes);

  // the base Jacobian
  mat_eye(Jbasecog);
  mat_eye(Jbasecogdes);

  Jbasecog[_Y_][N_CART+_X_] = -(cog.x[_Z_] - base_state.x[_Z_]);
  Jbasecog[_Z_][N_CART+_X_] =   cog.x[_Y_] - base_state.x[_Y_];
  
  Jbasecog[_X_][N_CART+_Y_] =   cog.x[_Z_] - base_state.x[_Z_];
  Jbasecog[_Z_][N_CART+_Y_] = -(cog.x[_X_] - base_state.x[_X_]);
  
  Jbasecog[_X_][N_CART+_Z_] = -(cog.x[_Y_] - base_state.x[_Y_]);
  Jbasecog[_Y_][N_CART+_Z_] =   cog.x[_X_] - base_state.x[_X_];
  
  
  
  Jbasecogdes[_Y_][N_CART+_X_] = -(cog_des.x[_Z_] - base_state.x[_Z_]);
  Jbasecogdes[_Z_][N_CART+_X_] =   cog_des.x[_Y_] - base_state.x[_Y_];
  
  Jbasecogdes[_X_][N_CART+_Y_] =   cog_des.x[_Z_] - base_state.x[_Z_];
  Jbasecogdes[_Z_][N_CART+_Y_] = -(cog_des.x[_X_] - base_state.x[_X_]);
  
  Jbasecogdes[_X_][N_CART+_Z_] = -(cog_des.x[_Y_] - base_state.x[_Y_]);
  Jbasecogdes[_Y_][N_CART+_Z_] =   cog_des.x[_X_] - base_state.x[_X_];

}

/*!*****************************************************************************
 *******************************************************************************
\note  compute_constraint_kinematics
\date  March 2006
   
\remarks 

      computes the constraint jacobian, its pseudo-inverse,  and its 
      null space projection from the given Jacobian of all endeffectors

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

     
 \param[in]     status  :  matrix of N_ENDEFFx6 elements of TRUE/FALSE to indicate
                           which constraints to use

 \param[in]     use_des :  use desired states instead of actual states, TRUE/FALSE
 \param[out]    Jc      :  constraint Jacobian
 \param[out]    Jcp     :  pseudo inverse of constraint Jacobian
 \param[out]    NJc     :  null space projection of contraint Jacobian


 ******************************************************************************/
#define RIDGE_FACTOR 1.e-6
void
compute_constraint_kinematics(int **status, int use_des, Matrix Jc, 
			      Matrix Jcp, Matrix NJc)
{

  int i,j,m,n;
  int count = 0;
  static int firsttime = TRUE;
  static Matrix JcJcT;
  static Matrix JcJcTinv;
  static Matrix Ec;
  SL_Cstate *cs;
  Matrix     JJ;
  
  if (firsttime) {
    firsttime = FALSE;
    JcJcT     = my_matrix(1,N_ENDEFFS*6,1,N_ENDEFFS*6);
    JcJcTinv  = my_matrix(1,N_ENDEFFS*6,1,N_ENDEFFS*6);
    Ec        = my_matrix(1,N_CART*2,1,N_CART*2);
    mat_eye(Ec);
  }

  if (use_des) {
    cs = cart_des_state;
    JJ = Jdes;
  } else {
    cs = cart_state;
    JJ = J;
  }

  // create the Jc matrix
  count = 0;
  for (i=1; i<=N_ENDEFFS; ++i) {

    // compute the Jacobian component due to the base: these results
    // come from cross productcs of the unit vectors of the base coordinate
    // with the endeffector-base position.
    Ec[_Y_][N_CART+_X_] = -(cs[i].x[_Z_] - base_state.x[_Z_]);
    Ec[_Z_][N_CART+_X_] =   cs[i].x[_Y_] - base_state.x[_Y_];

    Ec[_X_][N_CART+_Y_] =   cs[i].x[_Z_] - base_state.x[_Z_];
    Ec[_Z_][N_CART+_Y_] = -(cs[i].x[_X_] - base_state.x[_X_]);

    Ec[_X_][N_CART+_Z_] = -(cs[i].x[_Y_] - base_state.x[_Y_]);
    Ec[_Y_][N_CART+_Z_] =   cs[i].x[_X_] - base_state.x[_X_];

    for (j=1; j<=2*N_CART; ++j) {

      if (status[i][j]) {

	++count;

	for (m=1; m<=N_DOFS; ++m)
	  (Jc)[count][m] = JJ[(i-1)*2*N_CART+j][m];

	for (m=1; m<=2*N_CART; ++m)
	  (Jc)[count][N_DOFS+m] = Ec[j][m];	  

      }

    }
  }

  // compute pseudo-inverse
  mat_mult_outer_size(Jc,count,N_DOFS+2*N_CART,JcJcT);
  for (i=1; i<=count; ++i)
    JcJcT[i][i] += RIDGE_FACTOR;
  my_inv_ludcmp(JcJcT,count,JcJcTinv);
  mat_mult_transpose_normal_size(Jc,count,N_DOFS+2*N_CART,
				 JcJcTinv,count,count,
				 Jcp);

  // compute the null space projection matrix of Jc
  mat_mult_size(Jcp,N_DOFS+2*N_CART,count,
		Jc,count,N_DOFS+2*N_CART,
		NJc);
  for (i=1; i<=N_DOFS+2*N_CART; ++i)
    for (j=1; j<=N_DOFS+2*N_CART; ++j) {
      (NJc)[i][j] *= -1;
      if (i==j)
	(NJc)[i][j] += 1;
    }

#if 0
  // test pseudo inverse: this sould be an identity matrix
  MY_MATRIX(TEST,1,count,1,count);
  mat_mult_size(Jc,count,N_DOFS+6,Jcp,N_DOFS+6,count,TEST);
  print_mat("TEST",TEST);
#endif

}
 
/*!*****************************************************************************
 *******************************************************************************
\note  compute_constraint_cog_kinematics
\date  March 2006
   
\remarks 

      computes the constraint-weighted pseudo inverse of the cog jacobian, 
      and its  null space projection from the know COG Jacobian

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[in]     NJc        :  null space projection of contraints
 \param[in]     use_z      :  cog -x,y,z is controlled
                              FALSE: onle cog -x,y is controlled -- z is free
 \param[in]     status     :  matrix of N_ENDEFFx6 elements of TRUE/FALSE to indicate
                              which constraints to use
 \param[in]     use_des    :  use desired states TRUE/FALSE

 \param[out]    Jccog      :  the contraint cog Jacobian (i.e, Jcog*NJc)
 \param[out]    Jccogp     :  pseudo inverse of contraint cog Jacobian
 \param[out]    NcJccog    :  null space projection of Jtotcogp multiplied by
                              the constraint null space
 \param[out]    NJccog     :  null space projection of Jccog


 ******************************************************************************/
void
compute_constraint_cog_kinematics(Matrix NJc, int use_z, int **status, int use_des,
				  Matrix Jccog, 
				  Matrix Jccogp, Matrix NcJccog, 
				  Matrix NJccog) 
{

  int i,j,m,n,r;
  static int firsttime = TRUE;
  static Matrix uwJccog;
  static Matrix JccogJccogT;
  static Matrix JccogJccogTinv;
  static Matrix Ec;
  int    n_cart_real;

  if (firsttime) {
    firsttime           = FALSE;
    uwJccog             = my_matrix(1,N_CART,1,N_DOFS+2*N_CART);
    JccogJccogT         = my_matrix(1,N_CART,1,N_CART);
    JccogJccogTinv      = my_matrix(1,N_CART,1,N_CART);
    Ec                  = my_matrix(1,N_CART*2,1,N_CART*2);
    mat_eye(Ec);
  }

  if (use_z)
    n_cart_real = N_CART;
  else
    n_cart_real = N_CART-1;

  
  // compute the cog Jacobian component due to the base: these results
  // come from cross productcs of the unit vectors of the base coordinate
  // with the cog-base position.
  if (use_des) {
    Ec[_Y_][N_CART+_X_] = -(cog.x[_Z_] - base_state.x[_Z_]);
    Ec[_Z_][N_CART+_X_] =   cog.x[_Y_] - base_state.x[_Y_];
    
    Ec[_X_][N_CART+_Y_] =   cog.x[_Z_] - base_state.x[_Z_];
    Ec[_Z_][N_CART+_Y_] = -(cog.x[_X_] - base_state.x[_X_]);
    
    Ec[_X_][N_CART+_Z_] = -(cog.x[_Y_] - base_state.x[_Y_]);
    Ec[_Y_][N_CART+_Z_] =   cog.x[_X_] - base_state.x[_X_];
  } else {
    Ec[_Y_][N_CART+_X_] = -(cog_des.x[_Z_] - base_state.x[_Z_]);
    Ec[_Z_][N_CART+_X_] =   cog_des.x[_Y_] - base_state.x[_Y_];
    
    Ec[_X_][N_CART+_Y_] =   cog_des.x[_Z_] - base_state.x[_Z_];
    Ec[_Z_][N_CART+_Y_] = -(cog_des.x[_X_] - base_state.x[_X_]);
    
    Ec[_X_][N_CART+_Z_] = -(cog_des.x[_Y_] - base_state.x[_Y_]);
    Ec[_Y_][N_CART+_Z_] =   cog_des.x[_X_] - base_state.x[_X_];
  }
    
  // create the uwJccog matrix (unweighted Jccog matrix)
  for (i=1; i<=n_cart_real; ++i) {

    for (m=1; m<=N_DOFS; ++m) 
      if (use_des)
	uwJccog[i][m] = Jcogdes[i][m];
      else
	uwJccog[i][m] = Jcog[i][m];

    for (m=1; m<=2*N_CART; ++m)
      uwJccog[i][N_DOFS+m] = Ec[i][m];

  }

  // make sure the unconstraint endeffectors cannot be influenced by the cog controller by
  // modifying the uwJccog through zero-ing the columns of the swing leg
#if 0
  for (r=1; r<=N_ENDEFFS; ++r)
    //a simple check for unconstraint endeffectors
    if (status[r][_X_]==0 && status[r][_Y_]==0 && status[r][_Z_]== 0) {

      if (r==RIGHT_HAND) {

	for (i=R_SFE; i<=R_FING; ++i)
	  for (j=1; j<=n_cart_real; ++j)
	    uwJccog[j][i] = 0.0;

      } else if (r==LEFT_HAND) {

	for (i=L_SFE; i<=L_FING; ++i)
	  for (j=1; j<=n_cart_real; ++j)
	    uwJccog[j][i] = 0.0;

      } else if (r=RIGHT_FOOT) {

	for (i=R_FB; i<=R_AAA; ++i)
	  for (j=1; j<=n_cart_real; ++j)
	    uwJccog[j][i] = 0.0;

      } else if (r==LEFT_FOOT) {

	for (i=L_FB; i<=L_AAA; ++i)
	  for (j=1; j<=n_cart_real; ++j)
	    uwJccog[j][i] = 0.0;

      }
    }

  for (i=B_HR; i<=B_HN; ++i)
    for (j=1; j<=n_cart_real; ++j)
      uwJccog[j][i] = 0.0;

#endif

  // compute weighted pseudo-inverse
  mat_mult_size(uwJccog,n_cart_real,N_DOFS+2*N_CART,
		NJc,N_DOFS+2*N_CART,N_DOFS+2*N_CART,
		Jccog);
  mat_mult_outer_size(Jccog,n_cart_real,N_DOFS+2*N_CART,JccogJccogT);
  for (i=1; i<=n_cart_real; ++i)
    JccogJccogT[i][i] += RIDGE_FACTOR;
  my_inv_ludcmp(JccogJccogT,n_cart_real,JccogJccogTinv);
  mat_mult_transpose_normal_size(Jccog,n_cart_real,N_DOFS+2*N_CART,
				 JccogJccogTinv,n_cart_real,n_cart_real,
				 Jccogp);

  // make sure the _Z_ column is zero for !use_z
  if (!use_z)
    for (i=1; i<=N_DOFS+2*N_CART; ++i)
      (Jccogp)[i][_Z_] = 0.0;

  // compute the null space projection matrix of Jccog
  mat_mult_size(Jccogp,N_DOFS+2*N_CART,n_cart_real,
		Jccog,n_cart_real,N_DOFS+2*N_CART,
		NJccog);
  for (i=1; i<=N_DOFS+2*N_CART; ++i)
    for (j=1; j<=N_DOFS+2*N_CART; ++j) {
      (NJccog)[i][j] *= -1;
      if (i==j)
	(NJccog)[i][j] += 1;
    }
  mat_mult(NJc,NJccog,NcJccog);


#if 0
  // test pseudo inverse: this sould be an identity matrix
  MY_MATRIX(TEST,1,n_cart_real,1,n_cart_real);
  mat_mult_size(uwJccog,n_cart_real,N_DOFS+6,Jccogp,N_DOFS+6,n_cart_real,TEST);
  print_mat("TEST",TEST);
#endif


}
 
/*!*****************************************************************************
 *******************************************************************************
\note  compute_cog_kinematics
\date  March 2006
   
\remarks 

      A convenient wrapper to compute the pseudo inverse and the null
      space projection for cog balance control
      

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[in]     status     :  matrix of N_ENDEFFx6 elements of TRUE/FALSE to indicate
                              which constraints to use
 \param[in]     flag       :  use constraint null space projection, 
                              FALSE: use unconstraint null space projection
 \param[in]     use_z      :  cog -x,y,z is controlled
                              FALSE: onle cog -x,y is controlled -- z is free
 \param[in]     use_des    :  use desired states TRUE/FALSE
 \param[out]    Jccogp     :  pseudo inverse of contraint cog Jacobian  (N_DOFSxN_CART)
 \param[out]    NJccog     :  null space projection of Jccog  (N_DOFSxN_DOFS+2*N_CART)

 note: memory for all matrices needs to be provided


 ******************************************************************************/
void
compute_cog_kinematics(int **status, int flag, int use_z, int use_des,
		       Matrix Jccogp, Matrix NJccog)
{
  int i,j,r;

  MY_MATRIX(Jc_local,1,N_ENDEFFS*6,1,N_DOFS+2*N_CART);
  MY_MATRIX(Jcp_local,1,N_DOFS+2*N_CART,1,N_ENDEFFS*6);
  MY_MATRIX(NJc_local,1,N_DOFS+2*N_CART,1,N_DOFS+2*N_CART);

  MY_MATRIX(Jccog_local,1,N_CART,1,N_DOFS+2*N_CART);
  MY_MATRIX(Jccogp_local,1,N_DOFS+2*N_CART,1,N_CART);
  MY_MATRIX(NcJccog_local,1,N_DOFS+2*N_CART,1,N_DOFS+2*N_CART);
  MY_MATRIX(NJccog_local,1,N_DOFS+2*N_CART,1,N_DOFS+2*N_CART);


  // compute the null space of the contraints
  compute_constraint_kinematics(status, use_des, Jc_local, Jcp_local, NJc_local);

  compute_constraint_cog_kinematics(NJc_local, use_z, status, use_des, 
				    Jccog_local, Jccogp_local, 
				    NcJccog_local, NJccog_local);

  mat_equal_size(Jccogp_local,N_DOFS,N_CART,Jccogp);
  if (flag)
    mat_equal_size(NcJccog_local,N_DOFS,N_DOFS+2*N_CART,NJccog);
  else
    mat_equal_size(NJccog_local,N_DOFS,N_DOFS+2*N_CART,NJccog);


}
 
/*!*****************************************************************************
 *******************************************************************************
\note  compute_cog_velocity
\date  March 2006
   
\remarks 

      computes the cog velocity from the full cog jacobian (including
      base). It is assume that Jcog and the cog (position) have been 
      already computed.

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

     
     All results can be computed based on global variables. Results are
     assigned to the cog structure


 ******************************************************************************/
void
compute_cog_velocity(void)
{
  int i,j,r;
  double c[2*N_CART+1];
  static int firsttime = TRUE;
  static Matrix Jcogbase;
  
  if (firsttime) {
    firsttime = FALSE;
    Jcogbase = my_matrix(1,N_CART*2,1,N_CART*2);
    mat_eye(Jcogbase);
  }

  // the base Jacobian of the COG is almost an identity matrix, except for the
  // upper right quadrant
  for (j=1; j<=N_CART; ++j) {
    revoluteGJacColumn( cog.x,
			base_state.x,
			Jcogbase[j], // this is a equivalent to a unit vector
			c );
    for (r=1; r<=N_CART; ++r) {
      Jcogbase[r][N_CART+j] = c[r];
    }
  }

  for (i=1; i<=N_CART; ++i) {
    cog.xd[i] = 0;
    for (j=1; j<=N_DOFS; ++j)
      cog.xd[i] += Jcog[i][j]*joint_state[j].thd;
    for (j=1; j<=N_CART; ++j)
      cog.xd[i] += Jcogbase[i][j]*base_state.xd[j];
    for (j=1; j<=N_CART; ++j)
      cog.xd[i] += Jcogbase[i][N_CART+j]*base_orient.ad[j];
  }

}

/*!*****************************************************************************
 *******************************************************************************
\note  inverseDynamicsFloat
\date  Nov. 2008
\remarks 

          computes the inverse dynamics based on a projection into the
          controllable configuration space


 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[in]     dt : servo period
 \param[in]     status : matrix of which endeffectors are constraint N_ENDEFFS x 6 matrix
 \param[in]     use_des : use desired states instead of actual states, TRUE/FALSE
 \param[in,out] joint_ref_state   :
                       the reference acceleration in joint space 
                          (will be replaced with the projected version)
                       the reference velocity in joint space (only needed if xdd_ref 
                          or add_ref = NULL) 
                          (will be replaced with projected version if xdd_ref or add_ref = NULL)
                       uff is filled in as an additive term to existing uff
 \param[in]     xdd_ref : the base reference acceleration (pass NULL to create it)
 \param[in]     add_ref : the base reference angular acceleration (pass NULL to create it)
 \param[out]    fc : predicted contact forces in a matrix N_ENDEFFS x 6

 ******************************************************************************/
void 
inverseDynamicsFloat(double dt, iMatrix status, int use_des, 
		     SL_DJstate *joint_ref_state, double *xdd_ref, double *add_ref,
		     Matrix fc) 
{
  int i,j,n,m;
  int count=0;
  int create_full_state;
  int nu = 0;          // number of unconstraint DOFs
  int nf = N_DOFS+6;   // DOFs of full floating base space
  MY_MATRIX(rbdM,1,N_DOFS+6,1,N_DOFS+6);
  MY_VECTOR(rbdCG,1,N_DOFS+6);

  // external variables
  extern double servo_time;

  double aux;

  SL_DJstate  joint_tmp_state[N_DOFS+1];

  double xdd_mem[N_CART+1];
  double add_mem[N_CART+1];

  static int     firsttime = TRUE;
  static Matrix  last_Vu;
  static double  last_Vu_time;
  static iMatrix last_status;

  if (firsttime) {
    firsttime = FALSE;
    last_Vu = my_matrix(1,nf,1,nf);
    last_status = my_imatrix(1,N_ENDEFFS,1,2*N_CART);
    for (i=1; i<=N_ENDEFFS; ++i)
      for (j=1; j<=2*N_CART; ++j)
	last_status[i][j] = status[i][j];
  }

  // count the constraints
  for (i=1; i<=N_ENDEFFS; ++i)
    for (j=1; j<=N_CART*2; ++j) {
      if (status[i][j])
	++count;
      if (last_status[i][j] != status[i][j]) // this detects change of constraints
	last_Vu_time = 0;                    // this triggers reset of dVudt
      last_status[i][j] = status[i][j];
    }

  // the number of DOFs of the truely unconstraint space
  nu = nf - count;

  // lots of matrix macro for allocation
  MY_MATRIX(V,1,nf,1,nf);
  MY_MATRIX(U,1,count,1,nf);
  MY_VECTOR(s,1,nf);
  MY_MATRIX(S,1,N_DOFS,1,nf);
  mat_eye(S);
  MY_MATRIX(Vu,1,nf,1,nu);
  MY_MATRIX(dVudt,1,nf,1,nu);
  MY_MATRIX(Su,1,nu,1,nf);

  MY_MATRIX(SuVTST,1,nu,1,N_DOFS);
  MY_MATRIX(SuVTSTp,1,N_DOFS,1,nu);
  MY_MATRIX(AUX,1,nu,1,nu);
  MY_MATRIX(P,1,N_DOFS,1,nf);

  MY_MATRIX(SVu,1,N_DOFS,1,nu);

  MY_VECTOR(qdd_full_ref,1,nf);
  MY_VECTOR(v2,1,nf);
  MY_VECTOR(tmp,1,nf);
  MY_VECTOR(f,1,count);
  MY_VECTOR(qdd_ref,1,N_DOFS);
  MY_VECTOR(qd_ref,1,N_DOFS);

  MY_MATRIX(Jc,1,count,1,N_DOFS+2*N_CART);
  MY_MATRIX(Jcp,1,N_DOFS+2*N_CART,1,count);
  MY_MATRIX(NJc,1,N_DOFS+2*N_CART,1,N_DOFS+2*N_CART);

  MY_MATRIX(R,1,count,1,count);
  MY_MATRIX(fc_w,1,N_ENDEFFS,1,2*N_CART);
  MY_MATRIX(A,1,N_CART,1,N_CART);


  if (xdd_ref == NULL || add_ref == NULL) {
    create_full_state = TRUE;
    xdd_ref = xdd_mem;
    add_ref = add_mem;
  } else
    create_full_state = FALSE;

  for (i=1; i<=N_DOFS; ++i) {
    qdd_ref[i] = joint_ref_state[i].thdd;
    qd_ref[i]  = joint_ref_state[i].thd;
  }

  // get the constraint Jacobian
  compute_constraint_kinematics(status, use_des, Jc, Jcp, NJc);

  // compute an orthonormal basis of the constraint Jacobian  with SVD
  mat_equal_size(Jc,count,nf,U);
  my_svdcmp(U, count, nf, s, V);

  // sort the columns of U and V accoring to w coefficients
  for (j=1; j<=nf; ++j)
    for (i=1; i<=nf-j; ++i)
      if (s[i] < s[i+1]) {
	for (n=1; n<=nf; ++n) {
	  aux = V[n][i];
	  V[n][i] = V[n][i+1];
	  V[n][i+1] = aux;
	}
	for (n=1; n<=count; ++n) {
	  aux = U[n][i];
	  U[n][i] = U[n][i+1];
	  U[n][i+1] = aux;
	}
	aux = s[i];
	s[i]=s[i+1];
	s[i+1]=aux;
      }

  // determine Su (nu x nf)
  mat_zero(Su);
  for (i=1; i<=nu; ++i)
    Su[i][count+i] = 1.0;

  // Vu = V*SuT  (nf x nu)
  mat_mult_normal_transpose(V,Su,Vu);

  // -----------------------------------------------------------------------------------
  if (create_full_state) { /* only need if the full base state is computed from
			      joint base state only */

    // detect if this function was not called periodically
    if (fabs(last_Vu_time-(servo_time-dt)) > 0.1*dt)
      mat_equal_size(Vu,nf,nu,last_Vu);

    // project the previous Vu matrix onto the new Vu, and ortho-normalize again -- this
    // is needed to make sure that the derivative of this matrix does not jump
    reorthonormalizeMatrix(Vu, last_Vu, nf, nu, Vu);

    // compute time derivative of Vu
    for (i=1; i<=nf; ++i)
      for (j=1; j<=nu; ++j)
	dVudt[i][j] = (Vu[i][j]-last_Vu[i][j])/dt;
  
    // copy the Vu matrix into last_Vu
    mat_equal_size(Vu,nf,nu,last_Vu);
    last_Vu_time = servo_time;

    // SVu = S*Vu  (N_DOFS x nu)
    mat_mult_size(S,N_DOFS,nf,Vu,nf,nu,SVu);

  }


  // -----------------------------------------------------------------------------------

  // compute Su*V'*S' = Vu'*S' (nu x N_DOFS)
  mat_mult_transpose_transpose(Vu,S,SuVTST);

  // compute pseudo inverse of Su*V'*S' 
  mat_mult_outer(SuVTST,AUX);
  my_inv_ludcmp(AUX,nu,AUX);
  mat_mult_transpose_normal(SuVTST,AUX,SuVTSTp); // N_DOFS x nu

  // compute P = SuVTSTp*Su*VT = SuVTSTp*Vu'
  mat_mult_normal_transpose_size(SuVTSTp,N_DOFS,nu,Vu,nf,nu,P);

  // -----------------------------------------------------------------------------------
  if (create_full_state)
    computeFullFloatingBaseState(nu,Vu, dVudt, SVu, qdd_ref, qd_ref, xdd_ref, add_ref);
  // -----------------------------------------------------------------------------------

  // form the full float state vector
  for (i=1; i<=N_DOFS; ++i)
    qdd_full_ref[i] = qdd_ref[i];

  for (i=1; i<=N_CART; ++i)
    qdd_full_ref[N_DOFS+i] = xdd_ref[i];

  for (i=1; i<=N_CART; ++i)
    qdd_full_ref[N_DOFS+N_CART+i] = add_ref[i];

  // prepare computation of the unconstraint inverse dynamics with base
  for (i=1; i<=N_DOFS; ++i) {
    if (use_des) {
      joint_tmp_state[i].th  = joint_des_state[i].th;
      joint_tmp_state[i].thd = joint_des_state[i].thd;
    } else {
      joint_tmp_state[i].th  = joint_state[i].th;
      joint_tmp_state[i].thd = joint_state[i].thd;
    }
    joint_tmp_state[i].thdd  = qdd_ref[i];
    joint_tmp_state[i].uex = 0;
    joint_tmp_state[i].uff = 0;
  }

  for (i=1; i<=N_CART; ++i) 
    base_state.xdd[i] = xdd_ref[i];

  for (i=1; i<=N_CART; ++i)
    base_orient.add[i] = add_ref[i];
 
  double     fbase[2*N_CART+1];
  SL_InvDynNEBase(NULL, joint_tmp_state, endeff, &base_state, &base_orient, fbase);

  // eliminate viscous forces -- they will be added later again
  for (i=1; i<=nf; ++i) {
    SL_Jstate  jt;

    if (i<=N_DOFS) {
      jt.th = joint_tmp_state[i].th;
      jt.thd = joint_tmp_state[i].thd;
      v2[i] = joint_tmp_state[i].uff - compute_independent_joint_forces(jt, links[i]);
    } else 
      v2[i] = fbase[i-N_DOFS];
  }

  // and this full floating base command can be projected into joint
  // torques 
  for (i=1; i<=N_DOFS; ++i) {
    SL_Jstate  jt;
    jt.th = joint_tmp_state[i].th;
    jt.thd = joint_tmp_state[i].thd;
    for (j=1; j<=nf; ++j)
      joint_ref_state[i].uff += P[i][j] * v2[j];
    //joint_ref_state[i].uff += compute_independent_joint_forces(jt, links[i]);
  }

  // also compute the anticpated contact forces -----------------------------
  for (i=1; i<=N_DOFS; ++i)
    v2[i] -= joint_ref_state[i].uff;
  
  // this is the R matrix from the QR decomposition, derived from SVD, i.e.,
  // J = U S V'; J' = Q R; thus J' = V S U', and V = Q, S U' = R
  for (i=1; i<=count; ++i)
    for (j=1; j<=count; ++j) {
      R[i][j] = U[j][i]*s[i];
    }

  // invert this matirx
  my_inv_ludcmp(R,count,R);

  // compute tmp = V'*v2
  for (i=1; i<=nf; ++i) {
    tmp[i] = 0.0;
    for (j=1; j<=nf; ++j)
      tmp[i] += V[j][i]*v2[j];
  }

  // the contact forces are thus
  mat_vec_mult_size(R,count,count,tmp,count,f);

  // sort the forces into the appropriate matrix
  m = 0;
  for (i=1; i<=N_ENDEFFS; ++i)
    for (j=1; j<=2*N_CART; ++j)
      if (status[i][j])
	fc_w[i][j] = f[++m];
      else
	fc_w[i][j] = 0.0;

  // and convert the forces into leg coordinates
  for (i=1; i<=N_ENDEFFS; ++i) {
    if (use_des)
      mat_trans_size(Alink_des[link2endeffmap[i]],N_CART,N_CART,A);
    else
      mat_trans_size(Alink[link2endeffmap[i]],N_CART,N_CART,A);
    mat_vec_mult_size(A,N_CART,N_CART,fc_w[i],N_CART,fc[i]);
  }

  

}

/*!*****************************************************************************
 *******************************************************************************
\note  computeFullFloatingBaseAcceleration
\date  Nov. 2008
\remarks 

          computes the full floating base acceleration from only joint
          space acceleration and the constraint basis matrix and its
          derivative. 


 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[in]     nu          : the dimension of the unconstraint space
 \param[in]     Vu          : unconstraint space basis vector matrix
 \param[in]     dVudt       : time derivative of Vu
 \param[in]     SVu         : Vu projected into joint space
 \param[in,out] qdd_ref     : reference joint acceleration
 \param[in,out] qd_ref      : reference joint velocity
 \param[out]    xdd_ref     : reference base acceleration
 \param[out]    add_ref     : reference base angualar acceleration

 ******************************************************************************/
static void 
computeFullFloatingBaseState(int nu, Matrix Vu, Matrix dVudt, Matrix SVu, double *qdd_ref,
			     double *qd_ref, double *xdd_ref, double *add_ref)
{
  int i,j;
  int nf=N_DOFS+6;

  MY_MATRIX(AUX,1,nu,1,nu);
  MY_MATRIX(LS,1,nu,1,N_DOFS);
  MY_VECTOR(qd_full,1,nf);
  MY_VECTOR(qdd_full,1,nf);
  MY_VECTOR(v1,1,nu);
  MY_VECTOR(v2,1,N_DOFS);

  // We need to compute the desired reference acceleration at the 
  // floating base level, i.e., joint acceleration and base acceleration. This is
  // accomplished by first finding the least-squares solution of generating the
  // desired joint accelerations from the controllable coordinates, and then
  // using a projection of this acceleration to full floating base space.

  // compute the least squares solution in u space inv(Vu'*Vu)*Vu'
  mat_mult_transpose_normal(SVu,SVu,AUX);
  my_inv_ludcmp(AUX,nu,AUX);
  mat_mult_normal_transpose(AUX,SVu,LS);

  // the full state velocity from the least squares solution
  mat_vec_mult_size(LS,nu,N_DOFS,qd_ref,N_DOFS,v1);
  mat_vec_mult(Vu,v1,qd_full);

  for (i=1; i<=N_DOFS; ++i)
    qd_ref[i] = qd_full[i];

  // compute qdd - S dVudt Vu' qd_full
  for (i=1; i<=nu; ++i) {
    v1[i] = 0.0;
    for (j=1; j<=nf; ++j)
      v1[i] += Vu[j][i] * qd_full[j];
  }
  
  mat_vec_mult_size(dVudt,N_DOFS,nu,v1,nu,v2); // note that this includes S*dVudt
  for (i=1; i<=N_DOFS; ++i)
    v2[i] = qdd_ref[i] - v2[i];
  
  // multily with least squares matrices
  mat_vec_mult(LS,v2,v1);
  
  // compute dVudt'*qd_full_des and subtract from v1
  for (i=1; i<=nu; ++i) 
    for (j=1; j<=nf; ++j)
      v1[i] -= dVudt[j][i]*qd_full[j];
  
  // and finally get the desire acceleration in floating base space
  mat_vec_mult(Vu,v1,qdd_full);

  // sort the result back into appropriate arrays
  for (i=1; i<=N_DOFS; ++i)
    qdd_ref[i] = qdd_full[i];

  for (i=1; i<=N_CART; ++i)
    xdd_ref[i] = qdd_full[N_DOFS+i];

  for (i=1; i<=N_CART; ++i)
    add_ref[i] = qdd_full[N_DOFS+N_CART+i];

}

/*!*****************************************************************************
 *******************************************************************************
\note  reorthonormalizeMatrix
\date  Nov. 2008
\remarks 

     Given a matrix A and a matrix B, both with orthonormal columns, i.e.,
     A'A = I and B'B = I (eye is the size of the rank of A or B), a new
     matrix C is created of the size of A that is
     i) orthonormal in column space
     ii) has column vectors as close possible to the column vectors of B.

 *******************************************************************************
 Function Parameters: [in]=input,[out]=output

 \param[in]     A  : matrix to be re-orthormalized
 \param[in]     B  : template matrix for the directions of the column vectors of C
 \param[in]     nr : number of rows of matrices
 \param[in]     nc : number of columns of matrices
 \param[out]    C  : modified A matrix (note that memory for A and B can be the same)

 ******************************************************************************/
static void
reorthonormalizeMatrix( Matrix A, Matrix B, int nr, int nc, Matrix C)
{
  int i,j,n;
  double aux;
  double v[nr+1];
  MY_MATRIX(ATB,1,nc,1,nc);
  MY_MATRIX(Bp,1,nr,1,nc);

  // project B onto the range space of A: Bp  =  A*A'*B
  mat_mult_transpose_normal_size(A,nr,nc,B,nr,nc,ATB);
  mat_mult_size(A,nr,nc,ATB,nc,nc,Bp);

  // renormalize the Bp matrix to become C

  // start with the first column and just normalize
  aux = 0.0;
  for (i=1; i<=nr; ++i)
    aux += sqr(Bp[i][1]);
  aux = sqrt(aux);

  for (i=1; i<=nr; ++i)
    C[i][1] = Bp[i][1]/aux;

  // make sure that subsequent columns of C are always in the Null space of previous
  // columns
  for (i=2; i<=nc; ++i) {
    // project the current column vector on the range space of previous colums
    // and create component in null space
    for (j=1; j<=i-1; ++j) {
      v[j] = 0.0;
      for (n=1; n<=nr; ++n)
	v[j] += C[n][j]*Bp[n][i];
    }
    for (n=1; n<=nr; ++n) {
      C[n][i] = Bp[n][i];
      for (j=1; j<=i-1; ++j)
	C[n][i] -= v[j]*C[n][j];
    }
    // normalize new column
    aux = 0.0;
    for (n=1; n<=nr; ++n)
      aux += sqr(C[n][i]);
    aux = sqrt(aux);

    for (n=1; n<=nr; ++n)
      C[n][i] /= aux;
  }

}

